#include <16F630.h>
#fuses HS,NOWDT,MCLR,NOPROTECT,NOBROWNOUT
#use delay(clock=10000000)
#use rs232(baud=9600, xmit=PIN_C1, rcv=PIN_C2, INVERT)	//software UART
#use FAST_IO(C)
#use FAST_IO(A)

/*
 * CONSTANTS
 */
#define myAddr 0x0A			//This should be stored in EEPROM and be configurable
 							//But I ran out of time for a full implementation.
#define PWR PIN_C3			//These are hardware constants used
#define ECHO PIN_C4
#define SINIT PIN_C5
#define SDA PIN_C0
#define SCK PIN_A2

/*
 * We have a very limited functionality here, and the SEND is automatic.
 */
#define CMD_PING 0x01		//This tells the device to take a SONAR reading
#define CMD_SEND 0x81		//This tells the device to send the whatever data back
#define CMD_DEMO 0x0F		//puts the device in demo mode until power cycled.
/*
 * TRIS values
 */
#define ATRIS 0x04			//data direction for TRIS A
#define CTRIS 0x14			//data direction for TRIS C C0 output
#define RSPI  0x15			//turn C0 to input

/*
 * Communications states
 */
#define S_IDLE 0			// waiting for any clock
#define S_ADDRCMD 1			//Address and command input state
#define S_DOUT 2			//Sending data out
#define S_DIN 3				//Pulling in 16 bits of data
#define S_READY 4			//There is a command ready

/*
 * command interpretation patterns.  "locate" variables are "volatile"
 */
int cmdIn = 0;				//Most recent command received
#locate cmdIn = 0x5c		//Keep it in common memory for ISR use
int bCount = 0;				//bit coming in or going out
#locate bCount = 0x5b		//Kept in register memory for fast access in ISR
long dataIn = 0;
#locate dataIn = 0x59		//Should I ever take 16 bits of data in...
int spiState = 0;			//state machine variable for communications
#locate spiState = 0x58
int addrIn = 0;				//incoming address target
#locate addrIn = 0x57
long dataOut = 0;			//16 bit data to send back
#locate dataOut = 0x55

/*
 * save_w and save_status are at 5f, 5e and 5d which are available in ALL banks
 * all of these variables can be called "volatile" in that they are modified 
 * outside of the normal program's scope.
 */
int save_w;
#locate save_w=0x5f
int save_status;
#locate save_status=0x5e
int save_FSR;
#locate save_FSR=0x5d

/*
 * Give me direct access to a few SFR's that CCS doesn't give me.
 */
#byte INTCON = 0x0B
#byte FSR = 0x04
#byte status = 0x03

/*
 * Global values to be used by the program itself, not the ISR.
 */
int cmd = 0;											//The command received by master
long myData = 0;										//The data coming in or going out
#define SPI_TIMEOUT 250									//bit clock timeout threashold value

#int_global
void isr(void)
/*
 * I will handle the pre/post amble for the ISR because I know which memory locations I'll use.
 *
 * Incoming commands are read in, as would be any incoming data.  The main program
 * would determine what to do with the data.  Whenever a command comes in that has 
 * bit 7 (msb) of the command byte set the ISR will transmit back whatever is in the
 * data buffer.  The main program will have no choice in the matter, it had better have
 * the data in myData, because whatever IS there will be sent back! 
 */
{
	#asm
   	//store current state of processor
   	MOVWF save_w
   	SWAPF status,W
   	MOVWF save_status
   	SWAPF FSR,W
   	MOVWF save_FSR
   	BCF   status,5									//Set to page 0 for SFR's
	#endasm
	
	set_timer0(0);									//start the timeout clock ticking.
	switch (spiState)
	{
		case S_IDLE:								//Kick start the state machine
			bCount = 0;								//clear count
			spiState = S_ADDRCMD;					//and move right to the next state
		case S_ADDRCMD:
			if (bCount <8)							//We are getting the target address
			{
				shift_right(&addrIn,1,input(SDA));		//Get the current data bit
			}
			else if(bCount < 16)					//Now getting the command
			{
				shift_right(&cmdIn,1,input(SDA));
			}
			if (bCount == 15)						//we have an address and a command
			{
				if (addrIn == myAddr)				//This is for me
				{
					if ((cmdIn & 0x80) != 0)
					{
						spiState = S_DOUT;			//next state is data going out
						dataOut = myData;			//Get the current value to send
					}	
					else
					{
						spiState = S_DIN;			//next state is 16 bits data in
					}
				}	
			}		
			break;
			
		case S_DOUT:								//Next 16 bits go out
			set_tris_C(CTRIS);						//Make SDA an output now
			output_bit(SDA,shift_right(&dataOut,2,0));
		
			break;
			
		case S_DIN:									//Next 16 bits come in
			shift_right(&dataIn,2,input(SDA));
			
			break;
	
		case S_READY:							//Done, deal with data now.	
		default:								//Here just for a bug
		set_tris_C(RSPI);						//Make SDA an input now
		bCount--;								//Keep coming back to this state until reset	
		break;									//to S_READY
	}		
	bCount++;									//increment the bit count   
	if (bCount == 32)
	{
		spiState = S_READY;						//all input and output are done
	}
	
   	#asm
   	// restore processor and return from interrupt
   	BCF	  INTCON,1
   	SWAPF save_FSR,W
   	MOVWF FSR
   	SWAPF save_status,W
   	MOVWF status
   	SWAPF save_w,F
   	SWAPF save_w,W
   	#endasm  
}                                

void init(void)
/*
 * sets up everything before I use it.  Needed when using FAST_IO calls.
 * RC5 = Out : SONAR init
 * RC4 = In  : SONAR echo
 * RC3 = Out : SONAR ~pwr (pull low to power unit)
 * RC2 = In  : RxD (bit banged)
 * RC1 = Out : TxD (bit banged)
 * RC0 = <>  : SPI SDA 
 *
 * RA2 = In  : INT SPI SCK
 * RA1 = <>  : PGD
 * RA0 = <>  : PGC
 *
 * 10MHz clock, external ~MCLR
 */
{
	set_tris_C(RSPI);							//SDA is an input
	set_tris_A(ATRIS);

	//Turn off comparator
	setup_comparator(NC_NC_NC_NC);
	
	//Set up TMR1 to measure echo
	setup_timer_1(T1_INTERNAL | T1_DIV_BY_4);	//1.6us per tic, max 104.86ms
	
	setup_timer_0(RTCC_INTERNAL | RTCC_DIV_256); //6.5ms bit timeout for SPI clock
	
	enable_interrupts(INT_EXT);					//enable external INT (SCK line)
	enable_interrupts(GLOBAL);					//turn on the interrupts
}

int16 ping(void)
/*
 * ping() will do a single SONAR cycle
 * returns 16 bit value, 0 means timed out, any other number is range in
 * inches.
 *
 * SONAR is about 148us per inch round trip, TMR1 is 1.6us per tic.
 */
{
	int16 temp;								//holder for 16 bit time
	
	output_low(PWR);						//turn on power
	delay_ms(10);							//wait for power to come up
	set_timer1(0);							//clear the timer
	output_high(SINIT);						//send a ping
	while(!input_state(ECHO))
	{
		temp = get_timer1();
		if (temp > 65000)
		{
			temp = 0;
			break;
		}	
	}
	output_high(PWR);						//turn off power
	output_low(SINIT);						//de-assert init
	return (temp/92);						//148us/1.6us = 92.5, return inches
}

void demo(void)
/*
 * Just does repetitive pings as a demo.
 */
{
	long x;
	printf("%c%c",254,1);
	delay_ms(10);
	printf("SONAR demo");
	while(1)
	{
		x = ping();						//Get SONAR reading
		printf("%c%c",254,192);			//line 2
		printf("%lu inches  ",x);		//print the range
		delay_ms(5000);
	}		
}

void main(void)
{
	init();
	delay_ms(1000);
	//Printf commands are for a Scott Edwards 2x16 LCD.
	printf("%c%c",254,0x01);
	delay_ms(10);
	printf("Polaroid SONAR");
	printf("%c%c",254,192);
	printf("The eternal hack");
	delay_ms(1000);
		
	while(1)
	{
		/*
		 * Check for clock timeout, and recover.
		 */
		if ((spiSTATE != S_IDLE) && (get_timer0() > SPI_TIMEOUT))
		{
			if (spiState != S_READY)		//leave READY state alone
			{
				spiState = S_IDLE;			//transfer timeout, reset after 6.5ms
			}
		}	
		/*
		 * We have gotten a command, execute it and reset the state machine.
		 */
		if (spiState == S_READY)			//we have a command ready
		{
			cmd = cmdIn;					//get the command
			switch(cmd)
			{
				case CMD_PING:				//get a reading
					myData = ping();		//This data will go out when requested
					spiState = S_IDLE;		//ready to get another SPI command
					break;
					
				case CMD_SEND:				//We sent something back
					spiState = S_IDLE;		//reset the state machine
					break;
					
				case CMD_DEMO:				//go to permanent demo mode
					demo();					//and never return.
					break;
					
				default:
			}
		}	
	}	
}
