/*
* ATM168.c Code file to test working with ATMEGA168 talking to a PIC SONAR controller.
* You are free to use this any way you want.
* Using avr-gcc 4.3.0
*
*/

#undef F_CPU
#define F_CPU 16000000UL			//16MHz clock being used.
#define BAUD 38400L					//Bluesmirf Gold baud rate setting
#define BAUD_RR (F_CPU/16L/BAUD-1) 

#include <inttypes.h> 
#include <stdio.h>
#include <avr/io.h>
#include "atm168.h"

/*
 * Enable printf() functionality in avr-gcc libs
 */
FILE *fileDef;
int def_putc(char ch,FILE *fileDef);
int def_getc(FILE *fileDef);

/*
* This code exercises the SONAR board
*/
#define CMD_PING	0x01
#define CMD_SEND	0x81
#define CMD_DEMO	0x0F

/*
* Clock and data pins
*/
#define SDA PORTCbits.b2
#define SDA_IN PINCbits.b2
#define SCK PORTCbits.b1

/*
* Data Direction Register settings.
*/
#define SWITCH_C	0xFB			//Make PC2 an input
#define NORMAL_C	0xFF			//Make PC2 an output

int main(void)
{
	int data = 0;
	char addr = 0x0a;							//SPI device address I'm talking to
		
	initialize();								//Initialize the board

    fileDef = fdevopen(def_putc, def_getc);     //init stdio for printf() to work

	PORTBbits.b1 = 1;							//Turn on the debug LED so we can see it running.

	/*
	 * Just send out some example commands and get a SONAR reading back.
	 * It works!
	 */
	waitms(3000);
	printf("sending PING\n\r");
	waitms(100);
	sendSPI(addr,CMD_PING,0xA5A5);
	waitms(5000);
	printf("get reading\n\r");
	waitms(50);
	data = getSPI(addr,CMD_SEND);
	printf("read: %u\n\r",data);
	waitms(5000);
	printf("set SONAR to demo mode.\n\r");
	waitms(100);
	sendSPI(addr,CMD_DEMO,0);
	
	while(1)
	{
		PORTBbits.b1 = ~PORTBbits.b1;		//Blink an LED on this port to show we're running
		waitms(200);
	}
}

void initialize()
{
	DDRC = NORMAL_C; 						// All outputs for now...
	DDRB = 0xDE; 							// Set up the rest according to pinouts
	DDRD = 0xFE;

	ACSR = 0x80; 							// turn off the comparitor
	
	//Background timer
	TCCR2A = 0x00;							//No PWM
	TCCR2B = 0x05;							//prescale 128 for 16MHz clock
	
    //Enable UART (BAUD set above)
    UBRR0H = (unsigned char)(BAUD_RR >> 8);
    UBRR0L = (unsigned char)BAUD_RR;
    UCSR0B = (unsigned char)_BV(TXEN0) | _BV(RXEN0);
    UCSR0C = (unsigned char)_BV(UCSZ01) | _BV(UCSZ00);
    
	PORTD = 0x00;
	PORTB = 0x00;
	PORTC = 0x00;							//make sure SDA and SCK lines are low
}

/*
 * My special SPI(ish) command set
 */

void waste(void)
/*
 * This delays the clock for data setup timing
 */
{
	volatile int x;							//So compiler won't optimize away!
	
	for (x=0;x<100;x++)
		;
}

void sendSPI(char addr,char cmd,unsigned int data)
/*
 * Send an SPI command out.
 * SDA = PC0
 * SCK = PC1
 * 
 * My protocol starts with the clock low, then put data on SDA, then raise clock high.
 * Then send assume SDA is always an output.
 */
{
	unsigned int bCount = 0;							//bit count
	
	for (bCount = 0;bCount<8;bCount++)		//send the address
	{
		SDA = addr & 0x01;					//send bit 0
		waste();
		addr>>=1;							//get next bit
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		waitms(1);
	}
	for (bCount = 0;bCount<8;bCount++)		//send the command
	{
		SDA = cmd & 0x01;					//send bit 0
		waste();
		cmd>>=1;							//get next bit
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		waitms(1);
	}
	for (bCount = 0;bCount<16;bCount++)		//send the data, if any
	{
		SDA = data & 0x01;					//send bit 0
		waste();
		data>>=1;							//get next bit
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		waitms(1);
	}
	
}

unsigned int getSPI(char addr, char cmd)			
/*
 * get an SPI data packet
 * SDA = PC0
 * SCK = PC1
 * 
 * My protocol expects data to be good on falling edge of the SCK line.
 * get will flip the SDA line to be an input to get the 16 bit data back, then flip to output again.
 */
{
	unsigned int data=0;
	unsigned int bCount = 0;				//bit count
	unsigned int bit = 0x8000;				//set the msb
	
	for (bCount = 0;bCount<8;bCount++)		//send the address
	{
		SDA = addr & 0x01;					//send bit 0
		waste();
		addr>>=1;							//get next bit
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		waitms(1);
	}
	for (bCount = 0;bCount<8;bCount++)		//send the command
	{
		SDA = cmd & 0x01;					//send bit 0
		waste();
		cmd>>=1;							//get next bit
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		waitms(1);
	}
	
	DDRC = SWITCH_C;						//Now make SDA an input and get data.
	
	for (bCount = 0;bCount<16;bCount++)		//send the data, if any
	{
		SCK = 1;							//raise the clock
		waitms(1);							//keep the clock to 1KH or so
		SCK = 0;
		data>>=1;							//shift to next bit to fill
		if (SDA_IN == 1)
			data |= bit;					//set msb
		waitms(1);
	}
	
	DDRC = NORMAL_C;						//set back to all outputs
	SDA = 0;								//set data bit low too
	return(data);
}

void waitms(int d)
/*
 * This is tuned to 16MHz clock
 */
{
	int delay;

	TCNT2 = 130; 							//(255-130)*8us = about 1ms
	for (delay=0;delay<d;delay++)
	{
		while (TCNT2);
		TCNT2 = 130;
	}
}

/*
 * printf() functions use these.
 */
int def_putc(char ch,FILE *fileDef)
/*
 * The default putc routine for printf(
 */
 {
     while ((unsigned char)(UCSR0A & _BV(UDRE0)) == 0);

     UDR0 = ch;
     return ch;
 }

int def_getc(FILE *fileDef)
{
	//Crude, but workable
	if ((unsigned char)(UCSR0A & (1<<RXC0)))
		return (int)UDR0;
	else
		return (int)-1;
}

